/*
 *  $Id: asapcache.c,v 1.16 2002/08/22 09:37:22 dreibh Exp $
 *
 * RSerPool implementation.
 *
 * Realized in co-operation between Siemens AG
 * and University of Essen, Institute of Computer Networking Technology.
 *
 * Acknowledgement
 * This work was partially funded by the Bundesministerium fr Bildung und
 * Forschung (BMBF) of the Federal Republic of Germany (Frderkennzeichen 01AK045).
 * The authors alone are responsible for the contents.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * There are two mailinglists available at http://www.sctp.de/rserpool.html
 * which should be used for any discussion related to this implementation.
 *
 * Contact: rsplib-discussion@sctp.de
 *          dreibh@exp-math.uni-essen.de
 *
 * Purpose: ASAP Cache
 *
 */


#include "tdtypes.h"
#include "asapcache.h"
#include "utilities.h"
#include "asaperror.h"


/* ###### Print cache's content ########################################## */
static void asapCacheDump(struct ASAPInstance* asap, FILE* fd)
{
   fputs("*************** Name Cache Dump ***************\n",fd);
   fputs("\n",fd);
   poolNamespacePrint(asap->Cache,fd);
   fputs("***********************************************\n",fd);
}


/* ###### Update cache entry for pool element ############################ */
struct PoolElement* asapCacheUpdatePoolElement(struct ASAPInstance* asap,
                                               struct PoolHandle*   poolHandle,
                                               struct PoolElement*  poolElement,
                                               const bool           increment)
{
   struct PoolElement* copy;
   struct Pool*        pool;

   LOG_ACTION
   fprintf(stdlog,"Cache Update for ");
   poolElementPrint(poolElement,stdlog);
   LOG_END

   copy = poolNamespaceFindPoolElementByIdentifier(asap->Cache,poolElement->Identifier);
   if(copy != NULL) {
      LOG_ACTION
      fputs("Already cached, only timestamp update necessary.\n",stdlog);
      LOG_END

      copy->TimeStamp = getMicroTime();
      copy->Flags &= ~PEF_FAILED;
   }
   else {
      pool = poolNamespaceFindPool(asap->Cache,poolHandle);
      if(pool == NULL) {
         LOG_ACTION
         fputs("New pool, to be created first.\n",stdlog);
         LOG_END

         pool = poolNew(poolHandle,poolElement->Policy);
         if(pool != NULL) {
            poolNamespaceAddPool(asap->Cache,pool);
         }
      }

      if(pool != NULL) {
         LOG_ACTION
         fputs("Adding pool element.\n",stdlog);
         LOG_END

         copy = poolElementDuplicate(poolElement);
         if(copy != NULL) {
            copy->Handle = getPoolElementHandle();
            poolNamespaceAddPoolElement(asap->Cache,copy);
            poolAddPoolElement(pool,copy);
         }
      }
   }

   if((increment) && (copy != NULL))  {
      copy->UserCounter++;
   }

   LOG_VERBOSE2
   fputs("\n",stdlog);
   asapCacheDump(asap,stdlog);
   LOG_END
   return(copy);
}


/* ###### Update cache entries of pool ################################### */
void asapCacheUpdatePool(struct ASAPInstance* asap,
                         struct Pool*         pool)
{
   GList* poolElementList = g_list_first(pool->PoolElementList);
   while(poolElementList != NULL) {
      asapCacheUpdatePoolElement(asap,pool->Handle,(struct PoolElement*)poolElementList->data,false);
      poolElementList = g_list_next(poolElementList);
   }
}


/* ###### Purge pool element ############################################# */
void asapCachePurgePoolElement(struct ASAPInstance* asap,
                               struct PoolElement*  poolElement,
                               const bool           decrement)
{
   struct Pool*        pool;
   struct PoolElement* found = poolNamespaceFindPoolElementByIdentifier(
                                  asap->Cache,
                                  poolElement->Identifier);

   LOG_ACTION
   fprintf(stdlog,"Cache Purge for ");
   poolElementPrint(poolElement,stdlog);
   LOG_END

   if(found != NULL) {
      pool = found->OwnerPool;

      if(decrement) {
         if(found->UserCounter > 0) {
            found->UserCounter--;
         }
         else {
            LOG_WARNING
            fputs("Tried decrement of zero UserCounter!\n",stdlog);
            LOG_END
         }
      }

      if(found->UserCounter == 0) {
         poolNamespaceRemovePoolElement(asap->Cache,found);
         poolRemovePoolElement(pool,found);
         poolElementDelete(found);
         LOG_ACTION
         fputs("UserCounter=0 -> removing pool element.\n",stdlog);
         LOG_END

         if(pool->PoolElements == 0) {
            LOG_ACTION
            fputs("Pool is empty -> removing pool.\n",stdlog);
            LOG_END
            poolNamespaceRemovePool(asap->Cache,pool);
            poolDelete(pool);
         }
      }
      else {
         LOG_ACTION
         fprintf(stdlog,"UserCounter=%d -> pool element is still in use!\n",found->UserCounter);
         LOG_END
      }

      LOG_VERBOSE2
      fputs("\n",stdlog);
      asapCacheDump(asap,stdlog);
      LOG_END
   }
   else {
      LOG_ACTION
      fputs("Pool element not found!\n",stdlog);
      LOG_END
   }
}


/* ###### Purge pool ##################################################### */
void asapCachePurgePool(struct ASAPInstance* asap,
                        struct Pool*         pool)
{
   GList* poolElementList = g_list_first(pool->PoolElementList);
   while(poolElementList != NULL) {
      asapCachePurgePoolElement(asap,(struct PoolElement*)poolElementList->data,false);
      poolElementList = g_list_next(poolElementList);
   }
}


/* ###### Check for timeout of pool element ############################## */
static void asapCachePoolElementMaintenance(gpointer key, gpointer value, gpointer userData)
{
   struct ASAPInstance* asap        = (struct ASAPInstance*)userData;
   struct PoolElement*  poolElement = (struct PoolElement*)value;
   card64               now         = getMicroTime();

   if(poolElement->TimeStamp + asap->CacheElementTimeout <= now) {
      if(poolElement->UserCounter > 0) {
         LOG_ACTION
         fprintf(stdlog,"Cache timeout but UserCounter > 0 for ");
         poolElementPrint(poolElement,stdlog);
         LOG_END
      }
      else {
         asap->CachePurgeList = g_list_append(asap->CachePurgeList,poolElement);
      }
   }

   timerRestart(asap->CacheMaintenanceTimer,asap->CacheMaintenanceInterval);
}


/* ###### Check cache for timeouts and remove outdated pool elements ##### */
void asapCacheMaintenanceTimer(struct Dispatcher* dispatcher,
                               struct Timer*      timer,
                               void*              userData)
{
   struct ASAPInstance* asap = (struct ASAPInstance*)userData;
   GList*               list;

   LOG_ACTION
   fputs("Cache Maintenance...\n",stdlog);
   LOG_END

   dispatcherLock(asap->StateMachine);

   g_hash_table_foreach(asap->Cache->PoolElementsByHandle,asapCachePoolElementMaintenance,(void*)asap);

   if(asap->CachePurgeList) {
      list = g_list_first(asap->CachePurgeList);
      while(list != NULL) {
         asapCachePurgePoolElement(asap,(struct PoolElement*)list->data,false);
         list = g_list_next(list);
      }
      g_list_free(asap->CachePurgeList);
      asap->CachePurgeList = NULL;
   }

   dispatcherUnlock(asap->StateMachine);

   LOG_ACTION
   fputs("Cache Maintenance completed!\n",stdlog);
   LOG_END
}
